@@ -859,6 +859,41 @@ def add_command_subparser_args(
 __PARSER__ = None
 
 
+# ------------------------------------------------------------------------------------------------------------------------
+def program_defaults(
+    program:           Optional[str] = None,
+    version:           Optional[str] = None,
+    copyright_message: Optional[str] = None,
+) -> Tuple[str, str, str]:
+    """
+    Apply default values for program, version, and copyright message to be included in cli
+    application help text.
+
+    Parameters
+    ----------
+    program : str, optional
+        The name of the program. Defaults to `sys.argv[0]`
+    version : str
+        The current version of the program. Defaults to 'v?.?.?'
+    copyright_message : Optional[str]
+        The copyright message for the program. Defaults to '(C) copyright YEAR. All rights reserved)'
+
+    Returns
+    -------
+    Tuple[str, str, str]
+        The program, version, and copyright_message in a tuple in respective order.
+    """
+    year = datetime.datetime.utcnow().year
+    if copyright_message is None:
+        copyright_message = f"(C) Copyright {year}. All rights reserved."
+    if version is None:
+        version = "v?.?.?"
+    if program is None:
+        program = sys.argv[0]
+
+    return (program, version, copyright_message)
+
+
 # ------------------------------------------------------------------------------------------------------------------------
 def get_args_parser(
     program:           Optional[str] = None,
@@ -886,14 +921,10 @@ def get_args_parser(
     global __PARSER__
     if __PARSER__ is None:
 
-        # set defaults
-        year = datetime.datetime.utcnow().year
-        if copyright_message is None:
-            copyright_message = f"(C) Copyright {year}. All rights reserved."
-        if version is None:
-            version = "v?.?.?"
-        if program is None:
-            program = sys.argv[0]
+        # get the program default values
+        (program, version, copyright_message) = program_defaults(
+            program, version, copyright_message
+        )
 
         # initialize the main parser
         parser = argparse.ArgumentParser(
@@ -971,6 +1002,199 @@ def get_args_parser(
     return __PARSER__
 
 
+# ------------------------------------------------------------------------------------------------------------------------
+def convert_undefined_attributes_to_none( args: object ) -> None:
+    """
+    Change any `UNDEFINED` attributes to `None`.
+    This will be modified in-place, and nothing is returned.
+
+    Parameters
+    ----------
+    args : object
+        An arguments object to modify.
+    """
+    # Convert any undefined values to None
+    for (k, v) in args.__dict__.items():
+        if is_undefined(v):
+            args.__dict__[k] = None
+
+
+# ------------------------------------------------------------------------------------------------------------------------
+def apply_verbosity_args( args: object ) -> None:
+    """
+    Handle verbosity and quiet arguments, setting log verbosity accordingly.
+    The 'verbosity' and '__quiet' arguments (which were automatically added)
+    will be captured, applied, and removed from the `args` object, in-place.
+
+    Parameters
+    ----------
+    args : object
+        An arguments object.
+    """
+    # set log verbosity
+    verbosity = getattr(args, 'verbosity', UNDEFINED)
+    if is_defined(verbosity):  # pragma : no cover
+        delattr(args, 'verbosity')
+    verbosity = default_if_undefined(verbosity, 0)
+    verbosity = verbosity if (verbosity is not None) else 0
+
+    quiet     = getattr(args, '__quiet', UNDEFINED)
+    if is_defined(quiet):  # pragma : no cover
+        delattr(args, '__quiet')
+    quiet = default_if_undefined(quiet, 0)
+    quiet = quiet if (quiet is not None) else 0
+
+    set_verbosity( DEFAULT_VERBOSITY + verbosity - quiet )
+
+
+# ------------------------------------------------------------------------------------------------------------------------
+def get_args(
+    args:              Optional[Iterable[str]] = None,
+    program:           Optional[str] = None,
+    version:           Optional[str] = None,
+    copyright_message: Optional[str] = None,
+) -> Tuple[object, argparse.ArgumentParser]:
+    """
+    Get an argument parser based on all registered `command` functions, parse supplied
+    arguments, and return an arguments object.
+
+    Parameters
+    ----------
+    args: Iterable[str], optional
+        A list of positional arguments excluding the program name, defaults to `sys.argv[1:]`
+    program : str, optional
+        The name of the program. Defaults to `sys.argv[0]`
+    version : str
+        The current version of the program. Defaults to 'v?.?.?'
+    copyright_message : Optional[str]
+        The copyright message for the program. Defaults to '(C) copyright YEAR. All rights reserved)'
+
+
+    Returns
+    -------
+    Tuple[object, argparse.ArgumentParser]
+        A tuple whose first element is an arguments object ( as returned by
+        `argparse.ArgumentParser.parse_arguments` ) with attributes matching the registered
+        argument names for the selected command, and whose second element is the argument
+        parser itself.
+
+    """
+    # get the program default values
+    (program, version, copyright_message) = program_defaults(
+        program, version, copyright_message
+    )
+    if args is None:
+        args = sys.argv[1:]
+
+    # get parser
+    try:
+        parser = get_args_parser(
+            program = program,
+            version = version,
+            copyright_message = copyright_message
+        )
+    except Exception as e:
+        raise Exception("Error building parser(s).") from e
+
+    # parse args
+    try:
+        args = parser.parse_args( args = args )
+    except SystemExit:
+        raise
+    except Exception as e:  # noqa F722  # pragma : no cover
+        raise Exception("Error parsing arguments.") from e
+
+    # Convert any undefined values to None
+    convert_undefined_attributes_to_none(args)
+
+    # determine if we have a specified command
+    subparser = getattr(args, 'subparser', None)
+    if subparser is None:
+        logger.error("No command specified")
+        parser.print_help(file=sys.stderr)
+        sys.exit(-1)
+    delattr(args, 'subparser')
+
+    # done. return the result
+    return args, subparser
+
+
+# ------------------------------------------------------------------------------------------------------------------------
+def get_command_from_args(args: object) -> Tuple[Callable, str]:
+    """
+    Given a parsed arguments object, extract the command function and name, and
+    remove those attributes from the args object, in-place.
+
+    Parameters
+    ----------
+    args : object
+        An arguments object.
+
+    Returns
+    -------
+    Tuple[Callable, str]
+        The command function and command name.
+    """
+    command      = args.command_function
+    command_name = args.command_name
+    delattr(args, 'command_function')
+    delattr(args, 'command_name')
+    return (command, command_name)
+
+
+# ------------------------------------------------------------------------------------------------------------------------
+def handle_help_flag(args: object, parser: argparse.ArgumentParser) -> None:
+    """
+    Explicitly handle displaying of help information for a command when
+    the --help flag is specified. When present and True, the help is displayed
+    and the program exits. When present and False, the help attribute is
+    removed from args, in-place. When absent, no action is taken.
+
+    Parameters
+    ----------
+    args : object
+        An arguments object
+    parser : argparse.ArgumentParser
+        The argument parser.
+
+    """
+    help = getattr(args, 'help', None)
+    if help is not None:
+        if help:
+            parser.print_help(file = sys.stderr)
+            return 0
+        delattr(args, 'help')
+
+
+# ------------------------------------------------------------------------------------------------------------------------
+def display_invocation_details(command_name: str, kwargs: Dict[str, Any]) -> None:
+    """
+    Display program invocation details.
+
+    Parameters
+    ----------
+    command_name : str
+        The selected command name.
+    kwargs : Dict[str, Any]
+        Parsed dictionary of command line arguments.
+    """
+    logger.debug('=======================================================')
+    logger.debug('Command:   %s' % command_name)
+
+    if kwargs:
+        logger.debug('Arguments:')
+        mk = max(len(k) for k in kwargs) + 1
+        for (k, v) in kwargs.items():
+            try:
+                val = encode_json(v)
+            except:  # noqa F722
+                val = repr(v)
+            logger.debug(f"  {(k+':'):{mk}} {val}")
+    else:
+        logger.debug('Arguments: <<< No arguments provided >>>')
+    logger.debug('=======================================================')
+
+
 # ------------------------------------------------------------------------------------------------------------------------
 def run_command(
     args:              Optional[Iterable[str]] = None,
@@ -998,90 +1222,33 @@ def run_command(
         The command's exit code.
     """
 
-    year = datetime.datetime.utcnow().year
-    if copyright_message is None:
-        copyright_message = f"(C) Copyright {year}. All rights reserved."
-    if version is None:
-        version = "v?.?.?"
-    if program is None:
-        program = sys.argv[0]
-    if args is None:
-        args = sys.argv[1:]
-
-    # get the parser
     try:
-        parser = get_args_parser(
-            program = program,
-            version = version,
+        args, subparser = get_args(
+            args              = args,
+            program           = program,
+            version           = version,
             copyright_message = copyright_message
         )
-    except:  # noqa F722  # pragma : no cover
-        logger.exception("Error building parser(s).")
-        return -1
-
-    # parse the args
-    try:
-        args = parser.parse_args( args = args )
     except SystemExit:
         raise
     except:  # noqa F722  # pragma : no cover
-        logger.exception("Error parsing arguments.")
-        parser.print_help(file=sys.stderr)
+        logger.exception("Error getting program arguments")
         return -1
 
-    # Convert any undefined values to None
-    for (k, v) in args.__dict__.items():
-        if is_undefined(v):
-            args.__dict__[k] = None
-
-    # set log verbosity
-    verbosity = getattr(args, 'verbosity', UNDEFINED)
-    if is_defined(verbosity):  # pragma : no cover
-        delattr(args, 'verbosity')
-    verbosity = default_if_undefined(verbosity, 0)
-    verbosity = verbosity if (verbosity is not None) else 0
-
-    quiet     = getattr(args, '__quiet', UNDEFINED)
-    if is_defined(quiet):  # pragma : no cover
-        delattr(args, '__quiet')
-    quiet = default_if_undefined(quiet, 0)
-    quiet = quiet if (quiet is not None) else 0
-
-    set_verbosity( DEFAULT_VERBOSITY + verbosity - quiet )
-
-    # determine if we have a specified command
-    subparser = getattr(args, 'subparser', None)
-    if subparser is None:
-        logger.error("No command specified")
-        parser.print_help(file=sys.stderr)
-        return -1
-    delattr(args, 'subparser')
+    # apply logging verbosity args (verbose and quiet)
+    apply_verbosity_args(args)
 
     # get command details
-    command      = args.command_function
-    command_name = args.command_name
-    delattr(args, 'command_function')
-    delattr(args, 'command_name')
+    command, command_name = get_command_from_args(args)
 
     # get help flag
-    help = getattr(args, 'help', None)
-    if help is not None:
-        if help:
-            subparser.print_help(file = sys.stderr)
-            return 0
-        delattr(args, 'help')
+    handle_help_flag(args, subparser)
 
     # convert args to a dictionary
     kwargs = vars(args)
 
     # log command invocation arguments
-    logger.debug('=======================================================')
-    logger.debug('Command:   %s' % command_name)
-    logger.debug('Arguments:')
-    mk = max(len(k) for k in kwargs) + 1
-    for (k, v) in kwargs.items():
-        logger.debug(f"  {(k+':'):{mk}} {encode_json(v)}")
-    logger.debug('=======================================================')
+    display_invocation_details(command_name, kwargs)
 
     # invoke the command
     try:
